
# import ipykernel
# import urllib
# import json
import os
import shutil
from notebook import notebookapp
import nbformat
from requests import get

from .git_access import LocalRepo
from .git_access._run_cmd import run_cmd 

def notebook_to_pdf(repo=None,author=None,title=None):
    for srv in notebookapp.list_running_servers():
        try:
            if srv['token'] == '' and not srv['password']:
                server = srv['url'] + 'api/sessions'
            else:
                server = srv['url'] + 'api/sessions?token=' + srv['token']
            nb_server_info = get(server).json()[0]
        except:
            pass
        nb_loc = nb_server_info['path']
        if 'fileId=' in nb_loc: # Looks like we are in Colab
            nb_name = nb_server_info['name']
            from google.colab import files
            gdrive_home = '/content/drive/MyDrive'
            nb_path = gdrive_home + '/Colab Notebooks/'
            tmp_path = '/tmp'

            # If the drive is not already mounted, attempt to mount it
            if not os.path.isdir(gdrive_home):
                from google.colab import drive
                drive.mount('/content/drive')

            if 'fileId=https%3A%2F%2Fgithub.com%2F' in nb_loc: # and file on GitHub
                print('Note: Conversion will be performed on the most recent commit of this notebook on GitHub, not the working copy.')
                nb_name_us = nb_name.replace("%20-%20","_")
                nb_name_us = nb_name.replace("%20","_")
                nb_name_us = nb_name_us.replace("-","_")
                nb_name_us = nb_name_us.replace(":","_")
                nb_name_us = nb_name_us.replace("'","_")
                # nb_name = nb_name.replace("%20"," ")
                nb_name = nb_name_us
                if isinstance(repo,LocalRepo):
                    repo.pull()
                    shutil.copy(os.path.join(repo.repo_path, nb_name_us), os.path.join(tmp_path, nb_name))
                else:
                    print('Please pass a GitHub repo object as an argument.')
                    return
            else:
                if not os.path.isfile(os.path.join(nb_path, nb_name)):
                    raise ValueError(f"file '{nb_name}' not found in path '{nb_path}'")
                else:
                    shutil.copy(os.path.join(nb_path, nb_name), os.path.join(tmp_path, nb_name))
            


            # If PDF with the same name exists, remove it
            nb_file = os.path.join(tmp_path, nb_name)
            pdf_file = os.path.join(tmp_path, nb_name.split(".")[0] + ".pdf")
            if os.path.isfile(pdf_file):
                os.remove(pdf_file)

            # Add author(s) and title to notebook metadata (temp version only)
            tmp_nb = nbformat.read(nb_file, nbformat.NO_CONVERT)
            if title is not None:
                tmp_nb['metadata']['title'] = title
            if author is not None:
                if isinstance(author,str):
                    tmp_nb['metadata']['authors'] = [{"name": author}]
                elif isinstance(author,list):
                    author_list = []
                    for person in author:
                        author_list.append({"name": person})
                    tmp_nb['metadata']['authors'] = author_list
            nbformat.write(tmp_nb,nb_file)
            
            # Install the packages required for conversion
            print("Installing required packages. This often takes 1-2 minutes.")
            run_cmd("apt update >> /dev/null && apt install texlive-xetex texlive-fonts-recommended texlive-latex-recommended >> /dev/null")

            # Attempt to convert to PDF (via LaTeX)
            print(f"Preparing to convert '{nb_name}'")
            run_cmd(f"jupyter nbconvert --output-dir='{tmp_path}' '{nb_file}' --to pdf",verbose=True)

            # Attempt to download
            files.download(pdf_file)
        else:
            print('Sorry. Only implemented for Colab.')
            return

    


import numpy as np
import pandas as pd

def fitTable(fitParam, fitCov, paramLabels = ['slope', 'intercept'], ci_factor = 2):
    """
    fitTable

    Produces a pandas DataFrame with fit values and their uncertainties given
    array of fit parameters and covariance matrix.

    INPUTS:
        fitParam - a NumPy array with the fit parameters
        fitCov - covariance matrix from the fit algorithm

    KEYWORDS:
        paramLabels=['slope', 'intercept'] - list of strings, default assumes linear fit with slope first
        ci_factor=2 - confidence interval factor (1 = 68%, 2 = 95.4%, etc.)

    RETURNS:
        pandas DataFrame
    """
    return pd.DataFrame(data = [fitParam, ci_factor*np.sqrt(np.diagonal(fitCov))], columns = paramLabels, index = ['value', 'uncertainty'])

def prettyPolyFit(x, y, deg = 1, yerr = None, paramLabels = None, ci_factor = 2):
    """
    prettyPolyFit

    Performs a polynomial fit and returns in the results in a pandas DataFrame and a polynomial best fit function.

    INPUTS:
        x - NumPy array of independent values
        y - NumPy array of dependent values
        
    
    KEYWORDS:
        deg=1 - degree of polynomial fit to data (1 = line, 2 = quadratic)
        yerr=None - NumPy array of uncertainties in dependent values
        paramLabels=None - list of strings, if None it will be set to default dependent on value of deg
          If deg = 1 then paramLabels=['slope','intercept']
          If deg > 1 then paramLabels=['c0','c1','c2',...]
        ci_factor=2 - confidence interval factor (1 = 68%, 2 = 95.4%, etc.), only used if yerr=None. If yerr provided then confidence interval is same as those uncertainties.

    RETURNS:
        (fit_table, fit_func) where
            fit_table is a DataFrame with values and uncertainities of the fit parameters
            fit_func is polynomial function representing the best fit
    """
    if yerr is not None:
        ci_factor = 1
        fitPoly, fitCov = np.polyfit(x, y, deg, w=1/yerr, cov=True)
    else:
        fitPoly, fitCov = np.polyfit(x, y, deg, cov=True)
    if paramLabels is None:
        if deg == 1:
            paramLabels = ['slope','intercept']
        else:
            paramLabels = []
            for i in range(deg):
                paramLabels.append(f"c{i}")
    return (fitTable(fitPoly, fitCov, paramLabels=paramLabels, ci_factor=ci_factor), np.poly1d(fitPoly)) 


# __init__.py for whit-phys-utils

name = "whit_phys_utils"
from ._version import __version__

VERSION = __version__.split(".")

from .git_access import *
from .pretty_fit import fitTable, prettyPolyFit
from .create_pdf import notebook_to_pdf
from .dynamic_data_entry import DynamicDataEntry

__all__ = ["local_repository", "fitTable", "prettyPolyFit", "notebook_to_pdf","DynamicDataEntry"]


"""Run cmd in subprocess.run.

Some error logging is performed with goal of avoiding echo of secret info.
"""
import subprocess

def run_secret_cmd(cmd, verbose=False):
    """run_secret_cmd

    Runs a command using subprocess.run with some error logging, but no command echo to \
    avoid revealing secret info.
    """
    cp = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    if (verbose and cp.stdout):
        print(cp.stdout.strip())
    if (cp.returncode) or (verbose and cp.stderr):
        print(cp.stderr.strip())
    return cp.returncode

def run_cmd(cmd, verbose=False):
    """run_cmd

    Runs a command using subprocess.run with some error logging
    """
    cp = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    if (verbose and cp.stdout):
        print(cp.stdout.strip())
    if (cp.returncode) or (verbose and cp.stderr):
        print(cp.args.strip())
        print(cp.stderr.strip())
    return cp.returncode




from ._git_access import LocalRepo

def local_repository(repo, clone=True, branch="main", auth_method="env", expert_mode=False, verbose=False):
    """
    local_repository

    Call this method to create an instance of a GitHub repository in the 
    Google Colab environment. You can then use the class methods to push
    and pull from this repo, among other things, directly from Colab.

    INPUTS:
        repo - the link to the GitHub repository for cloning or the empty
                GitHub repo for initializing

    KEYWORDS:
        clone=True - indicates whether a new repository is being created
            locally and uploaded to GitHub, rather than a clone
        branch="main" - the specific branch to clone, if desired
        auth_method="env" - indicates where to look for your GitHub credentials.
            "env" will expect them in a .env or dotenv file on Google Drive, anything
            else will prompt the user to input them.
        expert_mode=False - indicates whether higher risk git commands are permitted
        verbose=False - indicates whether output of selected git commands should be displayed

    RETURNS:
        A LocalRepo instance
    """
    # Branch is forced to "main" for creating new repositories
    if not clone:
        branch = "main"

    return LocalRepo(repo, clone=clone, branch=branch, auth_method=auth_method, expert_mode = expert_mode, verbose=False)

__all__ = ["local_repository"]



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

exec(open("whit_phys_util/_version.py").read())

setuptools.setup(
    name="whit-phys-util",
    version=__version__,
    author="John Larkin",
    author_email="jlarkin@whitworth.edu",
    description="Tools to support use of Google Colab + GitHub Classroom for physics lab instruction",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JohnMLarkin/Colab-for-Physics-Lab-Tools",
    install_requires = [
        'ipydatagrid >= 1.3.2',
        'pydantic-settings >= 2.4.0'
    ],
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
)


import os
import sys
from getpass import getpass

from ._run_cmd import run_secret_cmd, run_cmd
from ._git_settings import git_settings

# Based on code developed by Jordan Lueck and Paddy Alton
# https://pypi.org/project/colab-repoclone/

class LocalRepo:
    def __init__(self, repo, clone=True, branch="main", auth_method="env", expert_mode=False, verbose=False):
        """
        __init__

        Gathers user's GitHub credentials and either clones or initializes
        desired repository.

        """

        if auth_method == "env":
            self.env_authentication()
        else:
            self.cli_authentication()

        repo_url = repo.split("//")[1]
        if clone:
            self.repo_dir = repo.split("/")[-1].replace(".git", "")
        else:
            self.repo_dir = input("Local Repository :: ")

        self.access_repo = f"https://{self.github_user}:{self.github_token}@{repo_url}"
        self.branch = branch
        self.expert_mode = expert_mode

        run_cmd(f"git config --global user.name {self.github_user}")
        run_cmd(f"git config --global user.email {self.github_email}")
        run_cmd("git config --global init.defaultBranch main")

        self.in_colab = "google.colab" in sys.modules
        if self.in_colab:
            self.base_dir = '/content'
        else:
            self.base_dir = os.getcwd()
        self.repo_path = os.path.join(self.base_dir, self.repo_dir)

        if clone:
            if os.path.exists(self.repo_path):
                print('Repository exists at this location. Pulling changes from GitHub.')
                self.pull(verbose)
            else:
                print('Cloning repository from GitHub.')
                self.clone(verbose)
        else:
            self.new()

    def clone(self, verbose=False):
        """
        clone

        Clones repository using username and access key.
        Set verbose=True to get output from clone command.
        """
        os.chdir(self.base_dir)
        if self.branch == "main":
            clone_cmd = f"git clone --progress {self.access_repo}"
        else:
            clone_cmd = f"git clone --progress --branch {self.branch} {self.access_repo}"
        run_secret_cmd(clone_cmd, verbose=verbose)
        os.chdir(self.repo_path)


    def new(self):
        """
        new

        Initializes a new repository in Google Colab environment using username and
        access key.

        """
        try:
            os.chdir(self.repo_path)
        except OSError:
            print(
                f"No directory named {self.repo_dir} exists. Are you sure you made it?"
            )
            return

        run_cmd("git init")
        origin = run_secret_cmd(f"git remote add origin {self.access_repo}", verbose=verbose)
        if origin:
            run_cmd("git remote rm origin")
            origin = run_secret_cmd(f"git remote add origin {self.access_repo}")
            if origin:
                print(
                    f"Command: < git remote add origin {self.access_repo} > failed. Check your permissions and that this repository exists on GitHub."
                )
                return


        ## First must pull (in case repo already exists), then push and set remote as upstream
        pull = run_cmd("git pull origin main --allow-unrelated-histories")
        if pull:
            print(
                "Command: < git pull origin main --allow-unrelated-histories > failed. Check your permissions."
            )
            return

        add = run_cmd("git add .")
        if add:
            print("Command: < git add . > failed. Check your permissions.")
            return

        commit = run_cmd("git commit -m 'First Commit from Google Colab'")
        if commit:
            print(
                f"Command: < git commit -m 'First Commit from Google Colab' > failed. Possibly because there were no files in /{self.repo_dir}"
            )
            return

        push = run_cmd("git push --set-upstream origin main")
        if push:
            print(
                "Command: < git push --set-upstream origin main > failed. Check your permissions."
            )


    def pull(self, verbose=False):
        """
        pull

        Pulls latest changes from GitHub repo into local Google Colab environment

        INPUTS:
            verbose=False - should output of git pull be displayed?

        """
        os.chdir(self.repo_path)
        pull = run_cmd("git pull", verbose=verbose)
        if pull:
            print("Command: < git pull > failed. Check your permissions.")

    def push(self, commit_msg=None, file_path=".", verbose=False):
        """
        push

        Commits and pushes latest changes to GitHub from Google Colab.

        KEYWORDS:
            commit_msg=None - message for this commit to GitHub. If none passed,
                will prompt for user input.
            file_path - path to specific files desired to push, defaults to all 
                files in repository.
            verbose=False - should output of git push be displayed?

        """
        if commit_msg is None:
            commit_msg = input("Commit Message :: ")

        check = """
        *************************************************************
        * Are you sure you want to push your changes?               *
        *                                                           *
        * Press "q" to abort. Press any other key to continue...    *
        *************************************************************
        """
        if input(check).lower() == "q":
            print("\n!! PUSH ABORTED !!\n")
            return

        os.chdir(self.repo_path)

        add = run_cmd(f"git add {file_path}", verbose=verbose)
        if add:
            print(f"Command: < git add {file_path} > failed. Check your permissions.")
            return

        commit = run_cmd(f"git commit -m '{commit_msg}'", verbose=verbose)
        if commit:
            print(
                f"Command: < git commit -m '{commit_msg}' > failed. Possibly because no changes were made. Also ensure there were no single or double quotation marks in your commit message."
            )
            return

        push = run_cmd("git push", verbose=verbose)
        if push:
            print("Command: < git push > failed. Check your permissions.")


    def new_branch(self, branch_name=None):
        """
        new_branch

        Creates a new branch off the current one and checks it out so future
        changes will be pushed to this new branch

        KEYWORDS:
            branch_name=None - the name of the new branch to create. If none
                passed, will prompt for user input.

        """
        if branch_name is None:
            branch_name = input("New Branch :: ")

        os.chdir(self.repo_path)

        brc = run_cmd(f"git branch {branch_name}")
        if brc:
            print(
                f"Command: < git branch {branch_name} > failed. Check your permissions."
            )
            return

        chk = run_cmd(f"git checkout {branch_name}")
        if chk:
            print(
                f"Command: < git checkout {branch_name} > failed. Check that this branch exists."
            )
            return

        self.branch = branch_name

        # Must push new branch to GitHub before making any changes to set the
        # upstream for future pushes from this branch
        push = run_cmd(f"git push --set-upstream origin {branch_name}")
        if push:
            print(
                f"Command: < git push --set-upstream origin {branch_name} > failed. Check your permissions."
            )


    def checkout(self, branch_name=None):
        """
        checkout

        Checks out an existing branch of the repository. All future pushes
        will push to this branch.

        KEYWORDS:
            branch_name=None - the name of the existing branch to checkout.
                If none passed, will prompt for user input.

        """

        if branch_name is None:
            branch_name = input("Checkout Branch :: ")

        os.chdir(self.repo_path)

        chk = run_cmd(f"git checkout {branch_name}")
        if chk:
            print(
                f"Command: < git checkout {branch_name} > failed. Check that this branch exists."
            )
            return

        self.branch = branch_name

    def reset(self, commit=None):
        """
        reset

        Performs a hard reset of the local repo to the specified commit,
        then force pushes this to rollback the repository, deleting all
        intermediate commits in the process.

        KEYWORDS:
            commit=None - the commit id to rollback to. If none passed,
                defaults to the previous commit.

        """
        check = """
        *****************************************************************
        *                         !! CAUTION !!                         *
        *                                                               *
        * Are you sure you want to rollback to a previous commit?       *
        *                                                               *
        * This is a hard reset, meaning all commits between the current *
        * and the one you are rolling back to will be lost.             *
        *                                                               *
        * Press "q" to abort. Press any other key to continue...        *
        *****************************************************************
        """
        if (not self.expert_mode):
            print("\n Repo reset is only permitted in expert mode \n")
            return

        if input(check).lower() == "q":
            print("\n!! RESET ABORTED !!\n")
            return

        os.chdir(self.repo_path)

        if commit is None:
            reset_cmd = "git reset --hard"
        else:
            reset_cmd = f"git reset --hard {commit}"

        reset = run_cmd(reset_cmd)
        if reset:
            print(
                f"Command: < {reset_cmd} > failed. Check the supplied commit id is a valid one."
            )

        push = run_cmd("git push --force")
        if push:
            print(f"Command: < git push --force > failed. Check your permissions.")


    def env_authentication(self):
        """
        env_authentication

        Checks for .env or .dotenv file with GitHub credentials. Used only when
        auth_method keyword passed to class instance is "env"

        """
        gs = git_settings()
        self.github_token = gs.gh_token
        self.github_user = gs.user_name
        self.github_email = gs.user_email
        if None in [self.github_token, self.github_user, self.github_email]:
            raise EnvironmentError(
                "Using auth_method='env', user_name, user_email, and gh_token must be \
                provided in .env or dotenv file"
            )

    def cli_authentication(self):
        """
        cli_authentication

        Uses getpass module to get user's GitHub credentials from standard input.
        Used only if auth_method keyword passed to class instance is *not* "env"

        """

        self.github_user = input("GitHub Username :: ")
        self.github_email = input("GitHub Email :: ")
        self.github_token = getpass("GitHub Authorization Token :: ")

